#
# Copyright (C) 2018-2019  UAVCAN Development Team  <uavcan.org>
# This software is distributed under the terms of the MIT License.
#

import sys
from .version import __version__, __license__
from . import _non_intrusive_test
from .namespace_parser import parse_namespace
from . import data_type
from . import parse_error


if sys.version_info[:2] < (3, 5):   # pragma: no cover
    print('A newer version of Python is required', file=sys.stderr)
    sys.exit(1)



#
# Copyright (C) 2018-2019  UAVCAN Development Team  <uavcan.org>
# This software is distributed under the terms of the MIT License.
#

import os
import typing
import tempfile
from .dsdl_parser import parse_definition, SemanticError, DSDLSyntaxError, ConfigurationOptions
from .dsdl_parser import UndefinedDataTypeError, AssertionCheckFailureError
from .dsdl_definition import DSDLDefinition, FileNameFormatError
from .data_type import CompoundType, StructureType, UnionType, ServiceType, ArrayType
from .namespace_parser import parse_namespace, FixedPortIDCollisionError, VersionsOfDifferentKindError
from .namespace_parser import MinorVersionsNotBitCompatibleError, MultipleDefinitionsUnderSameVersionError
from .namespace_parser import MinorVersionFixedPortIDError, NestedRootNamespaceError, NamespaceNameCollisionError


# Type annotation disabled here because MyPy is misbehaving, reporting these nonsensical error messages:
#   pydsdl/_non_intrusive_test.py:18: error: Missing type parameters for generic type
#   pydsdl/_non_intrusive_test.py: note: In function "_in_n_out":
#   pydsdl/_non_intrusive_test.py:18: error: Missing type parameters for generic type
_DIRECTORY = None       # type : typing.Optional[tempfile.TemporaryDirectory]


def _parse_definition(definition:         DSDLDefinition,
                      lookup_definitions: typing.Sequence[DSDLDefinition]) -> CompoundType:
    return parse_definition(definition,
                            lookup_definitions,
                            ConfigurationOptions())


def _define(rel_path: str, text: str) -> DSDLDefinition:
    assert _DIRECTORY
    path = os.path.join(_DIRECTORY.name, rel_path)
    os.makedirs(os.path.dirname(path), exist_ok=True)
    with open(path, 'w') as f:
        f.write(text)

    root_namespace_path = os.path.join(_DIRECTORY.name, rel_path.strip(os.sep).split(os.sep)[0])
    out = DSDLDefinition(path, root_namespace_path)
    print('New definition:', out, 'Root NS:', root_namespace_path)
    return out


def _in_n_out(test: typing.Callable[[], None]) -> typing.Callable[[], None]:
    def decorator() -> None:
        global _DIRECTORY
        _DIRECTORY = tempfile.TemporaryDirectory()
        try:
            test()
        finally:
            _DIRECTORY = None   # Preserving the contents for future inspection if needed

    return decorator


@_in_n_out
def _unittest_define() -> None:
    # I DON'T ALWAYS WRITE UNIT TESTS
    d = _define('uavcan/test/65000.Message.1.2.uavcan', '# empty')
    assert _DIRECTORY is not None
    assert d.full_name == 'uavcan.test.Message'
    assert d.version == (1, 2)
    assert d.fixed_port_id == 65000
    assert d.file_path == os.path.join(_DIRECTORY.name, 'uavcan/test/65000.Message.1.2.uavcan')
    assert open(d.file_path).read() == '# empty'

    # BUT WHEN I DO, I WRITE UNIT TESTS FOR MY UNIT TESTS
    d = _define('uavcan/Service.255.254.uavcan', '# empty 2')
    assert d.full_name == 'uavcan.Service'
    assert d.version == (255, 254)
    assert d.fixed_port_id is None
    assert d.file_path == os.path.join(_DIRECTORY.name, 'uavcan/Service.255.254.uavcan')
    assert open(d.file_path).read() == '# empty 2'


@_in_n_out
def _unittest_simple() -> None:
    abc = _define(
        'vendor/nested/29000.Abc.1.2.uavcan',
        '''
        @deprecated
        uint8 CHARACTER = '#'
        int8 a
        truncated int64[<33] b
        '''
    )
    assert abc.fixed_port_id == 29000
    assert abc.full_name == 'vendor.nested.Abc'
    assert abc.version == (1, 2)

    p = _parse_definition(abc, [])
    print('Parsed:', p)
    assert isinstance(p, StructureType)
    assert p.full_name == 'vendor.nested.Abc'
    assert p.source_file_path.endswith('vendor/nested/29000.Abc.1.2.uavcan')
    assert p.source_file_path == abc.file_path
    assert p.fixed_port_id == 29000
    assert p.deprecated
    assert p.version == (1, 2)
    assert p.bit_length_range == (14, 14 + 64 * 32)
    assert len(p.attributes) == 3
    assert len(p.fields) == 2
    assert str(p.fields[0].data_type) == 'saturated int8'
    assert p.fields[0].name == 'a'
    assert str(p.fields[1].data_type) == 'truncated int64[<=32]'      # Note: normalized representation
    assert p.fields[1].name == 'b'
    assert len(p.constants) == 1
    assert str(p.constants[0].data_type) == 'saturated uint8'
    assert p.constants[0].name == 'CHARACTER'
    assert p.constants[0].initialization_expression == "'#'"
    assert isinstance(p.constants[0].value, int)
    assert p.constants[0].value == ord('#')

    t = p.fields[1].data_type
    assert isinstance(t, ArrayType)
    assert str(t.element_type) == 'truncated int64'

    empty_new = _define(
        'vendor/nested/Empty.255.255.uavcan',
        ''''''
    )

    empty_old = _define(
        'vendor/nested/Empty.255.254.uavcan',
        ''''''
    )

    constants = _define(
        'another/Constants.5.0.uavcan',
        '''
        float64 PI = 3.1415926535897932384626433
        '''
    )

    service = _define(
        'another/300.Service.0.1.uavcan',
        '''
        @union
        @deprecated
        vendor.nested.Empty.255.255 new_empty_implicit
        vendor.nested.Empty.255.255 new_empty_explicit
        vendor.nested.Empty.255.254 old_empty
        ---#---#---#---#---#---#---#---#---
        Constants.5.0 constants      # RELATIVE REFERENCE
        vendor.nested.Abc.1.2 abc
        '''
    )

    p = _parse_definition(service, [
        abc,
        empty_new,
        empty_old,
        constants,
    ])
    print('Parsed:', p)
    assert isinstance(p, ServiceType)
    assert p.full_name == 'another.Service'
    assert p.fixed_port_id == 300
    assert p.deprecated
    assert p.version == (0, 1)
    assert not p.constants

    assert len(p.fields) == 2
    assert p.fields[0].name == 'request'
    assert p.fields[1].name == 'response'
    req, res = [x.data_type for x in p.fields]
    assert isinstance(req, UnionType)
    assert isinstance(res, StructureType)
    assert req.full_name == 'another.Service.Request'
    assert res.full_name == 'another.Service.Response'
    assert req is p.request_type
    assert res is p.response_type

    assert len(req.constants) == 0
    assert len(req.fields) == 3
    assert req.number_of_variants == 3
    assert req.deprecated
    assert not req.has_fixed_port_id
    assert req.version == (0, 1)
    assert req.bit_length_range == (2, 2)   # Remember this is a union
    assert [x.name for x in req.fields] == ['new_empty_implicit', 'new_empty_explicit', 'old_empty']

    t = req.fields[0].data_type
    assert isinstance(t, StructureType)
    assert t.full_name == 'vendor.nested.Empty'
    assert t.version == (255, 255)          # Selected implicitly

    t = req.fields[1].data_type
    assert isinstance(t, StructureType)
    assert t.full_name == 'vendor.nested.Empty'
    assert t.version == (255, 255)          # Selected explicitly

    t = req.fields[2].data_type
    assert isinstance(t, StructureType)
    assert t.full_name == 'vendor.nested.Empty'
    assert t.version == (255, 254)          # Selected explicitly

    assert len(res.constants) == 0
    assert len(res.fields) == 2
    assert res.deprecated
    assert not res.has_fixed_port_id
    assert res.version == (0, 1)
    assert res.bit_length_range == (14, 14 + 64 * 32)

    t = res.fields[0].data_type
    assert isinstance(t, StructureType)
    assert t.full_name == 'another.Constants'
    assert t.version == (5, 0)

    t = res.fields[1].data_type
    assert isinstance(t, StructureType)
    assert t.full_name == 'vendor.nested.Abc'
    assert t.version == (1, 2)

    union = _define(
        'another/Union.5.9.uavcan',
        '''
        @union
        truncated float16 PI = 3.1415926535897932384626433
        uint8 a
        vendor.nested.Empty.255.255[5] b
        truncated bool [ <= 255 ] c
        '''
    )

    p = _parse_definition(union, [
        empty_old,
        empty_new,
    ])

    assert p.full_name == 'another.Union'
    assert p.version == (5, 9)
    assert p.fixed_port_id is None
    assert not p.has_fixed_port_id
    assert not p.deprecated
    assert isinstance(p, UnionType)
    assert p.number_of_variants == 3
    assert len(p.constants) == 1
    assert p.constants[0].name == 'PI'
    assert str(p.constants[0].data_type) == 'truncated float16'
    assert p.bit_length_range == (2, 2 + 8 + 255)
    assert len(p.fields) == 3
    assert str(p.fields[0]) == 'saturated uint8 a'
    assert str(p.fields[1]) == 'vendor.nested.Empty.255.255[5] b'
    assert str(p.fields[2]) == 'truncated bool[<=255] c'


@_in_n_out
def _unittest_error() -> None:
    from pytest import raises

    def standalone(rel_path: str, definition: str, allow_unregulated: bool=False) -> CompoundType:
        cfg = ConfigurationOptions()
        cfg.allow_unregulated_fixed_port_id = allow_unregulated
        return parse_definition(_define(rel_path, definition), [], cfg)

    with raises(SemanticError, match='(?i).*port ID.*'):
        standalone('vendor/10000.InvalidRegulatedSubjectID.1.0.uavcan', 'uint2 value')

    with raises(SemanticError, match='(?i).*port ID.*'):
        standalone('vendor/10.InvalidRegulatedServiceID.1.0.uavcan', 'uint2 v1\n---\nint64 v2')

    with raises(SemanticError, match='(?i).*subject ID.*'):
        standalone('vendor/100000.InvalidRegulatedSubjectID.1.0.uavcan', 'uint2 value')

    with raises(SemanticError, match='(?i).*service ID.*'):
        standalone('vendor/1000.InvalidRegulatedServiceID.1.0.uavcan', 'uint2 v1\n---\nint64 v2')

    with raises(SemanticError, match='(?i).*multiple attributes under the same name.*'):
        standalone('vendor/AttributeNameCollision.1.0.uavcan', 'uint2 value\nint64 value')

    with raises(SemanticError, match='(?i).*tagged union cannot contain less than.*'):
        standalone('vendor/SmallUnion.1.0.uavcan', '@union\nuint2 value')

    assert standalone('vendor/invalid_constant_value/A.1.0.uavcan',
                      'bool BOOLEAN = false').constants[0].name == 'BOOLEAN'
    with raises(SemanticError, match='.*Invalid value for boolean constant.*'):
        standalone('vendor/invalid_constant_value/A.1.0.uavcan', 'bool BOOLEAN = 0')   # Should be false

    with raises(SemanticError, match='(?i).*could not evaluate.*'):
        standalone('vendor/invalid_constant_value/A.1.0.uavcan', 'bool BOOLEAN = undefined_identifier')

    with raises(DSDLSyntaxError):
        standalone('vendor/invalid_constant_value/A.1.0.uavcan', 'bool BOOLEAN = -')

    with raises(SemanticError, match='(?i).*exceeds the range.*'):
        standalone('vendor/invalid_constant_value/A.1.0.uavcan', 'uint10 INTEGRAL = 2000')

    with raises(SemanticError, match='(?i).*character.*'):
        standalone('vendor/invalid_constant_value/A.1.0.uavcan', "uint8 CH = '\u0451'")

    with raises(SemanticError, match='.*uint8.*'):
        standalone('vendor/invalid_constant_value/A.1.0.uavcan', "uint9 CH = 'q'")

    with raises(SemanticError, match='.*uint8.*'):
        standalone('vendor/invalid_constant_value/A.1.0.uavcan', "int8 CH = 'q'")

    with raises(SemanticError, match='(?i).*type.*'):
        standalone('vendor/invalid_constant_value/A.1.0.uavcan', "int8 CH = 1.0")

    with raises(SemanticError, match='(?i).*type.*'):
        standalone('vendor/invalid_constant_value/A.1.0.uavcan', "float32 CH = true")

    with raises(SemanticError, match='(?i).*type.*'):
        standalone('vendor/invalid_constant_value/A.1.0.uavcan', "float32 CH = 't'")

    with raises(DSDLSyntaxError):
        standalone('vendor/syntax_error/A.1.0.uavcan', 'bool array[10]')

    with raises(SemanticError, match='(?i).*array size.*'):
        standalone('vendor/array_size/A.1.0.uavcan', 'bool[0] array')

    with raises(SemanticError, match='(?i).*array size.*'):
        standalone('vendor/array_size/A.1.0.uavcan', 'bool[<1] array')

    with raises(SemanticError, match='(?i).*service response marker.*'):
        standalone('vendor/service/A.1.0.uavcan', 'bool request\n---\nbool response\n---\nbool again')

    with raises(SemanticError, match='(?i).*unknown directive.*'):
        standalone('vendor/directive/A.1.0.uavcan', '@sho_tse_take')

    with raises(SemanticError, match='(?i).*requires an expression.*'):
        standalone('vendor/directive/A.1.0.uavcan', '@assert')

    with raises(SemanticError, match='(?i).*does not expect an expression.*'):
        standalone('vendor/directive/A.1.0.uavcan', '@union worker')

    with raises(SemanticError, match='(?i).*version number.*'):
        standalone('vendor/version/A.0.0.uavcan', '')

    with raises(SemanticError, match='(?i).*version number.*'):
        standalone('vendor/version/A.0.256.uavcan', '')

    with raises(FileNameFormatError):
        standalone('vendor/version/A.0..256.uavcan', '')

    with raises(SemanticError, match='(?i).*version number.*'):
        standalone('vendor/version/A.256.0.uavcan', '')

    with raises(SemanticError, match='(?i).*cannot be specified for compound.*'):
        standalone('vendor/types/A.1.0.uavcan', 'truncated uavcan.node.Heartbeat.1.0 field')

    with raises(UndefinedDataTypeError, match='(?i).*no type named.*'):
        standalone('vendor/types/A.1.0.uavcan', 'nonexistent.TypeName.1.0 field')

    with raises(UndefinedDataTypeError, match='(?i).*no suitable major version'):
        _parse_definition(
            _define('vendor/types/A.1.0.uavcan', 'ns.Type_.1.0 field'),
            [
                _define('ns/Type_.2.0.uavcan', ''),
            ]
        )

    with raises(UndefinedDataTypeError, match='(?i).*no suitable minor version'):
        _parse_definition(
            _define('vendor/types/A.1.0.uavcan', 'ns.Type_.1.0 field'),
            [
                _define('ns/Type_.2.0.uavcan', ''),
                _define('ns/Type_.1.1.uavcan', ''),
            ]
        )

    with raises(DSDLSyntaxError, match='(?i).*Invalid type declaration.*'):
        _parse_definition(
            _define('vendor/types/A.1.0.uavcan', 'int128 field'),
            [
                _define('ns/Type_.2.0.uavcan', ''),
                _define('ns/Type_.1.1.uavcan', ''),
            ]
        )

    with raises(SemanticError, match='(?i).*type.*'):
        _parse_definition(
            _define('vendor/invalid_constant_value/A.1.0.uavcan', 'ns.Type_.1.1 VALUE = 123'),
            [
                _define('ns/Type_.2.0.uavcan', ''),
                _define('ns/Type_.1.1.uavcan', ''),
            ]
        )

    with raises(UndefinedDataTypeError):
        defs = [
            _define('vendor/circular_dependency/A.1.0.uavcan', 'B.1.0 b'),
            _define('vendor/circular_dependency/B.1.0.uavcan', 'A.1.0 b'),
        ]
        _parse_definition(defs[0], defs)

    with raises(SemanticError, match='(?i).*union directive.*'):
        _parse_definition(
            _define('vendor/misplaced_directive/A.1.0.uavcan', 'ns.Type_.2.0 field\n@union'),
            [
                _define('ns/Type_.2.0.uavcan', ''),
            ]
        )

    with raises(SemanticError, match='(?i).*deprecated directive.*'):
        _parse_definition(
            _define('vendor/misplaced_directive/A.1.0.uavcan', 'ns.Type_.2.0 field\n@deprecated'),
            [
                _define('ns/Type_.2.0.uavcan', ''),
            ]
        )

    with raises(SemanticError, match='(?i).*deprecated directive.*'):
        _parse_definition(
            _define('vendor/misplaced_directive/A.1.0.uavcan', 'ns.Type_.2.0 field\n---\n@deprecated'),
            [
                _define('ns/Type_.2.0.uavcan', ''),
            ]
        )


@_in_n_out
def _unittest_print() -> None:
    printed_items = None    # type: typing.Optional[typing.Tuple[DSDLDefinition, int, typing.Any]]

    def print_handler(definition: DSDLDefinition, line_number: int, value: typing.Any) -> None:
        nonlocal printed_items
        printed_items = definition, line_number, value

    config = ConfigurationOptions()
    config.print_handler = print_handler

    parse_definition(
        _define(
            'ns/A.1.0.uavcan',
            '''# line number 1
            # line number 2
            @print 2 + 2 == 4   # line number 3
            # line number 4
            '''),
        [],
        config
    )
    assert printed_items
    assert printed_items[0].full_name == 'ns.A'
    assert printed_items[1] == 3
    assert printed_items[2]

    parse_definition(_define('ns/B.1.0.uavcan', '@print false'), [], config)
    assert printed_items
    assert printed_items[0].full_name == 'ns.B'
    assert printed_items[1] == 1
    assert not printed_items[2]

    parse_definition(
        _define(
            'ns/Offset.1.0.uavcan',
            '''@print _offset_    # Not recorded
            uint8 a
            @print _offset_
            '''),
        [],
        config
    )
    assert printed_items
    assert printed_items[0].full_name == 'ns.Offset'
    assert printed_items[1] == 3
    assert printed_items[2] == {8}

    # The nested type has the following set: {2, 10, 18}.
    # We can have up to two elements of that type, so what we get can be expressed graphically as follows:
    #    A   B | +
    # ---------+------
    #    2   2 |  4
    #   10   2 | 12
    #   18   2 | 20
    #    2  10 | 12
    #   10  10 | 20
    #   18  10 | 28
    #    2  18 | 20
    #   10  18 | 28
    #   18  18 | 36
    # If we were to remove duplicates, we end up with: {4, 12, 20, 28, 36}
    parse_definition(
        _define(
            'ns/ComplexOffset.1.0.uavcan',
            '''
            Array.1.0[2] bar
            @print _offset_
            '''),
        [
            _define('ns/Array.1.0.uavcan', 'uint8[<=2] foo')
        ],
        config
    )
    assert printed_items
    assert printed_items[0].full_name == 'ns.ComplexOffset'
    assert printed_items[1] == 3
    assert printed_items[2] == {4, 12, 20, 28, 36}


@_in_n_out
def _unittest_assert() -> None:
    from pytest import raises

    _parse_definition(
        _define(
            'ns/A.1.0.uavcan',
            '''
            @assert _offset_ == {0}
            @assert _offset_.min == _offset_.max
            Array.1.0[2] bar
            @assert _offset_ == {4, 12, 20, 28, 36}
            @assert _offset_.min == 4
            @assert _offset_.max == 36
            @assert _offset_ % 4 == {0}
            @assert _offset_ % 8 == {4}
            @assert _offset_ % 10 == {4, 2, 0, 8, 6}
            @assert _offset_ * 2 == {8, 24, 40, 56, 72}
            @assert 2 * _offset_ == {8, 24, 40, 56, 72}
            @assert _offset_ / 4 == {1, 3, 5, 7, 9}
            @assert _offset_ - 4 == {0, 8, 16, 24, 32}
            @assert _offset_ + 4 == {8, 16, 24, 32, 40}
            uint64 big
            @assert _offset_ - 64 == {4, 12, 20, 28, 36}
            @assert _offset_.min == 68
            @assert _offset_.max == 100  # 36 + 64
            @assert _offset_.max <= 100
            @assert _offset_.max < 101
            @assert _offset_ <= 100
            @assert _offset_ < 101
            @assert _offset_ >= 68
            @assert _offset_ > 67
            @assert _offset_ == _offset_
            '''),
        [
            _define('ns/Array.1.0.uavcan', 'uint8[<=2] foo')
        ]
    )

    with raises(SemanticError, match='(?i).*invalid operand.*'):
        _parse_definition(
            _define(
                'ns/B.1.0.uavcan',
                '''
                uint64 big
                @assert _offset_ == {64}
                @assert _offset_ + 1.0 == {64}
                '''),
            []
        )

    with raises(SemanticError, match='(?i).*cannot be compared.*'):
        _parse_definition(
            _define(
                'ns/C.1.0.uavcan',
                '''
                uint64 big
                @assert _offset_ == 64
                '''),
            []
        )

    _parse_definition(
        _define(
            'ns/D.1.0.uavcan',
            '''
            @union
            float32 a
            uint64 b
            @assert _offset_ == {33, 65}
            '''),
        []
    )

    _parse_definition(
        _define(
            'ns/E.1.0.uavcan',
            '''
            @union
            uint8 A = 0
            float32 a
            uint8 B = 1
            uint64 b
            uint8 C = 2
            @assert _offset_ == {33, 65}
            uint8 D = 3
            '''),
        []
    )

    with raises(SemanticError, match='(?i).*unions.*'):
        _parse_definition(
            _define(
                'ns/F.1.0.uavcan',
                '''
                @union
                @assert _offset_.min == 33
                float32 a
                uint64 b
                @assert _offset_ == {33, 65}
                '''),
            []
        )

    with raises(AssertionCheckFailureError):
        _parse_definition(
            _define(
                'ns/G.1.0.uavcan',
                '''
                float32 a
                @assert _offset_.min == 8
                '''),
            []
        )

    with raises(SemanticError, match='(?i).*yield a boolean.*'):
        _parse_definition(
            _define(
                'ns/H.1.0.uavcan',
                '''
                float32 a
                @assert _offset_.min
                '''),
            []
        )


def _unittest_parse_namespace() -> None:
    from pytest import raises
    import tempfile
    directory = tempfile.TemporaryDirectory()

    def _define(rel_path: str, text: str) -> None:
        path = os.path.join(directory.name, rel_path)
        os.makedirs(os.path.dirname(path), exist_ok=True)
        with open(path, 'w') as f:
            f.write(text)

    _define(
        'zubax/First.1.0.uavcan',
        """
        uint8[<256] a
        @assert _offset_.min == 8
        @assert _offset_.max == 2048
        """
    )

    _define(
        'zubax/29001.Message.1.0.uavcan',
        """
        void6
        zubax.First.1.0[<=2] a
        @assert _offset_.min == 8
        @assert _offset_.max == 4104
        """
    )

    _define(
        'zubax/nested/300.Spartans.30.0.uavcan',
        """
        @deprecated
        @union
        float16 small
        float32 just_right
        float64 woah
        ---
        """
    )

    _define('zubax/nested/300.Spartans.30.0.txt', 'completely unrelated stuff')
    _define('zubax/300.Spartans.30.0', 'completely unrelated stuff')

    parsed = parse_namespace(
        os.path.join(directory.name, 'zubax'),
        []
    )
    print(parsed)
    assert len(parsed) == 3
    assert 'zubax.First' in [x.full_name for x in parsed]
    assert 'zubax.Message' in [x.full_name for x in parsed]
    assert 'zubax.nested.Spartans' in [x.full_name for x in parsed]

    _define(
        'zubax/colliding/300.Iceberg.30.0.uavcan',
        """
        ---
        """
    )

    with raises(FixedPortIDCollisionError):
        parse_namespace(
            os.path.join(directory.name, 'zubax'),
            []
        )


def _unittest_parse_namespace_versioning() -> None:
    from pytest import raises
    import tempfile
    import glob
    directory = tempfile.TemporaryDirectory()

    def _define(rel_path: str, text: str) -> None:
        path = os.path.join(directory.name, rel_path)
        os.makedirs(os.path.dirname(path), exist_ok=True)
        with open(path, 'w') as f:
            f.write(text)

    def _undefine_glob(rel_path_glob: str) -> None:
        path = os.path.join(directory.name, rel_path_glob)
        for g in glob.glob(path):
            os.remove(g)

    _define(
        'ns/Spartans.30.0.uavcan',
        """
        @deprecated
        @union
        float16 small
        float32 just_right
        float64 woah
        ---
        """
    )

    _define(
        'ns/Spartans.30.1.uavcan',
        """
        @deprecated
        @union
        uint16 small
        int32 just_right
        float64[1] woah
        ---
        """
    )

    parsed = parse_namespace(
        os.path.join(directory.name, 'ns'),
        []
    )
    print(parsed)
    assert len(parsed) == 2

    _define(
        'ns/Spartans.30.2.uavcan',
        """
        @deprecated
        @union
        uint16 small
        int32 just_right
        float64[<=1] woah
        ---
        """
    )

    with raises(MinorVersionsNotBitCompatibleError):
        parse_namespace(
            os.path.join(directory.name, 'ns'),
            []
        )

    _define(
        'ns/Spartans.30.2.uavcan',
        """
        @deprecated
        @union
        uint16 small
        int32 just_right
        float64[1] woah
        """
    )

    with raises(VersionsOfDifferentKindError):
        parse_namespace(
            os.path.join(directory.name, 'ns'),
            []
        )

    _undefine_glob('ns/Spartans.30.[01].uavcan')

    _define(
        'ns/Spartans.30.0.uavcan',
        """
        @deprecated
        @union
        uint16 small
        float32 just_right
        float64[1] woah
        """
    )

    parsed = parse_namespace(
        os.path.join(directory.name, 'ns'),
        []
    )
    print(parsed)
    assert len(parsed) == 2

    _define(
        'ns/Spartans.30.1.uavcan',
        """
        @deprecated
        @union
        uint16 small
        float32 just_right
        float64[<=1] woah
        """
    )

    with raises(MinorVersionsNotBitCompatibleError):
        parse_namespace(
            os.path.join(directory.name, 'ns'),
            []
        )

    _define(
        'ns/Spartans.30.1.uavcan',
        """
        @deprecated
        @union
        uint16 small
        float32 just_right
        int64 woah
        """
    )

    _define(
        'ns/28700.Spartans.30.2.uavcan',
        """
        @deprecated
        @union
        uint16 small
        int32 just_right
        float64[1] woah
        """
    )

    with raises(MultipleDefinitionsUnderSameVersionError):
        parse_namespace(os.path.join(directory.name, 'ns'), [])

    _undefine_glob('ns/Spartans.30.2.uavcan')

    parsed = parse_namespace(
        os.path.join(directory.name, 'ns'),
        []
    )
    assert len(parsed) == 3

    _undefine_glob('ns/Spartans.30.0.uavcan')
    _define(
        'ns/28700.Spartans.30.0.uavcan',
        """
        @deprecated
        @union
        uint16 small
        float32 just_right
        float64[1] woah
        """
    )

    with raises(MinorVersionFixedPortIDError):
        parse_namespace(os.path.join(directory.name, 'ns'), [])

    _undefine_glob('ns/Spartans.30.1.uavcan')
    _define(
        'ns/28700.Spartans.30.1.uavcan',
        """
        @deprecated
        @union
        uint16 small
        float32 just_right
        float64[1] woah
        """
    )

    parsed = parse_namespace(
        os.path.join(directory.name, 'ns'),
        []
    )
    assert len(parsed) == 3

    _undefine_glob('ns/28700.Spartans.30.1.uavcan')
    _define(
        'ns/28701.Spartans.30.1.uavcan',
        """
        @deprecated
        @union
        uint16 small
        float32 just_right
        float64[1] woah
        """
    )

    with raises(MinorVersionFixedPortIDError):
        parse_namespace(os.path.join(directory.name, 'ns'), [])

    # Adding new major version under the same FPID
    _undefine_glob('ns/28701.Spartans.30.1.uavcan')
    _define(
        'ns/28700.Spartans.31.0.uavcan',
        """
        @deprecated
        @union
        uint16 small
        float32 just_right
        float64[1] woah
        """
    )

    with raises(FixedPortIDCollisionError):
        parse_namespace(os.path.join(directory.name, 'ns'), [])

    # Major version zero allows us to re-use the same FPID under a different (non-zero) major version
    _undefine_glob('ns/28700.Spartans.31.0.uavcan')
    _define(
        'ns/28700.Spartans.0.1.uavcan',
        """
        @deprecated
        @union
        uint16 small
        float32 just_right
        float64[1] woah
        """
    )

    parsed = parse_namespace(os.path.join(directory.name, 'ns'), [])     # no error
    assert len(parsed) == 3


def _unittest_parse_namespace_faults() -> None:
    try:
        parse_namespace('/foo/bar/baz', ['/bat/wot', '/foo/bar/baz/bad'])
    except NestedRootNamespaceError as ex:
        print(ex)
    else:               # pragma: no cover
        assert False

    try:
        parse_namespace('/foo/bar/baz', ['/foo/bar/zoo', '/foo/bar/doo/roo/baz'])
    except NamespaceNameCollisionError as ex:
        print(ex)
    else:               # pragma: no cover
        assert False
    try:
        parse_namespace('/foo/bar/baz', ['/foo/bar/zoo', '/foo/bar/doo/roo/zoo', '/foo/bar/doo/roo/baz'])
    except NamespaceNameCollisionError as ex:
        print(ex)
    else:               # pragma: no cover
        assert False


@_in_n_out
def _unittest_inconsistent_deprecation() -> None:
    from pytest import raises

    _parse_definition(
        _define('ns/A.1.0.uavcan', ''),
        [
            _define('ns/B.1.0.uavcan',
                    '''
                    @deprecated
                    A.1.0 a
                    ''')
        ]
    )

    with raises(SemanticError, match='(?i).*depend.*deprecated.*'):
        _parse_definition(
            _define(
                'ns/C.1.0.uavcan',
                '''
                X.1.0 b
                '''
            ),
            [
                _define('ns/X.1.0.uavcan', '@deprecated')
            ]
        )

    _parse_definition(
        _define('ns/D.1.0.uavcan',
                '''
                @deprecated
                X.1.0 b
                '''),
        [
            _define('ns/X.1.0.uavcan', '@deprecated')
        ]
    )



#
# Copyright (C) 2018-2019  UAVCAN Development Team  <uavcan.org>
# This software is distributed under the terms of the MIT License.
#

import os
import typing
import logging
import fnmatch
from collections import defaultdict
from .data_type import CompoundType, ServiceType
from .dsdl_definition import DSDLDefinition
from .dsdl_parser import parse_definition, ConfigurationOptions, PrintDirectiveOutputHandler
from .parse_error import ParseError, InternalError, InvalidDefinitionError


DSDL_FILE_GLOB = '*.uavcan'


_logger = logging.getLogger(__name__)
_LOG_LIST_ITEM_PREFIX = ' ' * 4


class NamespaceNameCollisionError(InvalidDefinitionError):
    """
    Raised when there is more than one namespace under the same name.
    This may occur if there are identically named namespaces located in different directories.
    """
    pass


class NestedRootNamespaceError(InvalidDefinitionError):
    """
    Nested root namespaces are not allowed. This exception is thrown when this rule is violated.
    """
    pass


class FixedPortIDCollisionError(InvalidDefinitionError):
    """
    Raised when there is more than one data type, or different major versions of the same data type
    using the same fixed port ID.
    """
    pass


class MinorVersionsNotBitCompatibleError(InvalidDefinitionError):
    """
    Raised when definitions under the same major version are not bit-compatible.
    """
    pass


class MultipleDefinitionsUnderSameVersionError(InvalidDefinitionError):
    """
    For example:
        Type.1.0.uavcan
        58000.Type.1.0.uavcan
        58001.Type.1.0.uavcan
    """
    pass


class VersionsOfDifferentKindError(InvalidDefinitionError):
    """
    Definitions that share the same name but are of different kinds.
    """
    pass


class MinorVersionFixedPortIDError(InvalidDefinitionError):
    """
    Different fixed port ID under the same major version, or a fixed port ID was removed under the same
    major version.
    """
    pass


def parse_namespace(root_namespace_directory:        str,
                    lookup_directories:              typing.Iterable[str],
                    print_directive_output_handler:  typing.Optional[PrintDirectiveOutputHandler]=None,
                    allow_unregulated_fixed_port_id: bool=False,
                    skip_assertion_checks:           bool=False) -> \
        typing.List[CompoundType]:
    """
    Parse all DSDL definitions in the specified root namespace directory.

    :param root_namespace_directory:    The path of the root namespace directory that will be parsed.
                                        For example, "dsdl/uavcan" to parse the "uavcan" namespace.

    :param lookup_directories:          List of other namespace directories containing data type definitions that are
                                        referred to from the parsed root namespace. For example, if you are parsing a
                                        vendor-specific namespace, the list of lookup directories should always include
                                        a path to the standard root namespace "uavcan", otherwise the types defined in
                                        the vendor-specific namespace won't be able to use data types from the standard
                                        namespace.

    :param print_directive_output_handler:  If provided, this callable will be invoked when a @print directive is
                                            encountered. If not provided, print directives will not produce any output
                                            except for the log (at the INFO level).

    :param allow_unregulated_fixed_port_id: Do not reject unregulated fixed port identifiers.
                                            This is a dangerous feature that must not be used unless you understand the
                                            risks. The background information is provided in the UAVCAN specification.

    :param skip_assertion_checks:         Do not evaluate assertion check expressions. This option can be used to
                                          accelerate parsing significantly. Use sparingly.

    :return: A list of CompoundType.

    :raises: ParseError, OSError (if directories do not exist or inaccessible)
    """
    # Add the own root namespace to the set of lookup directories, remove duplicates
    lookup_directories = list(set(list(lookup_directories) + [root_namespace_directory]))

    # Normalize paths
    root_namespace_directory = os.path.abspath(root_namespace_directory)
    lookup_directories = list(map(lambda d: str(os.path.abspath(d)), lookup_directories))
    _logger.debug('Lookup directories are listed below:')
    for a in lookup_directories:
        _logger.debug(_LOG_LIST_ITEM_PREFIX + a)

    # Check the namespaces
    _ensure_no_nested_root_namespaces(lookup_directories)
    _ensure_no_namespace_name_collisions(lookup_directories)

    # Construct DSDL definitions from the target and the lookup dirs
    target_dsdl_definitions = _construct_dsdl_definitions_from_namespace(root_namespace_directory)
    _logger.debug('Target DSDL definitions are listed below:')
    for x in target_dsdl_definitions:
        _logger.debug(_LOG_LIST_ITEM_PREFIX + str(x))

    lookup_dsdl_definitions = []    # type: typing.List[DSDLDefinition]
    for ld in lookup_directories:
        lookup_dsdl_definitions += _construct_dsdl_definitions_from_namespace(ld)

    _logger.debug('Lookup DSDL definitions are listed below:')
    for x in lookup_dsdl_definitions:
        _logger.debug(_LOG_LIST_ITEM_PREFIX + str(x))

    # Parse the constructed definitions
    configuration_options = ConfigurationOptions()
    configuration_options.print_handler = print_directive_output_handler
    configuration_options.allow_unregulated_fixed_port_id = bool(allow_unregulated_fixed_port_id)
    configuration_options.skip_assertion_checks = bool(skip_assertion_checks)

    types = _parse_namespace_definitions(target_dsdl_definitions,
                                         lookup_dsdl_definitions,
                                         configuration_options)

    # Note that we check for collisions in the parsed namespace only.
    # We intentionally ignore (do not check for) possible collisions in the lookup directories,
    # because that would exceed the expected scope of responsibility of the parser, and the lookup
    # directories may contain issues and mistakes that are outside of the control of the user (e.g.,
    # they could be managed by a third party) -- the user shouldn't be affected by mistakes committed
    # by the third party.
    _ensure_no_fixed_port_id_collisions(types)
    _ensure_minor_version_compatibility(types)

    return types


def _parse_namespace_definitions(target_definitions:    typing.List[DSDLDefinition],
                                 lookup_definitions:    typing.List[DSDLDefinition],
                                 configuration_options: ConfigurationOptions) -> \
        typing.List[CompoundType]:
    """
    Construct type descriptors from the specified target definitions.
    Allow the target definitions to use the lookup definitions within themselves.
    :param target_definitions:  Which definitions to parse.
    :param lookup_definitions:  Which definitions can be used by the parsed definitions.
    :return: A list of types.
    """
    types = []  # type: typing.List[CompoundType]
    for tdd in target_definitions:
        try:
            parsed = parse_definition(tdd,
                                      lookup_definitions,
                                      configuration_options=configuration_options)
        except ParseError as ex:    # pragma: no cover
            ex.set_error_location_if_unknown(path=tdd.file_path)
            raise ex
        except Exception as ex:     # pragma: no cover
            raise InternalError(culprit=ex, path=tdd.file_path) from ex
        else:
            types.append(parsed)

    return types


def _ensure_no_fixed_port_id_collisions(types: typing.List[CompoundType]) -> None:
    for a in types:
        for b in types:
            different_names = a.full_name != b.full_name
            different_major_versions = a.version.major != b.version.major
            # Must be the same kind because port ID sets of subjects and services are orthogonal
            same_kind = isinstance(a, ServiceType) == isinstance(b, ServiceType)
            # Data types where the major version is zero are allowed to collide
            both_released = (a.version.major > 0) and (b.version.major > 0)

            fpid_must_be_different = same_kind and (different_names or (different_major_versions and both_released))

            if fpid_must_be_different:
                if a.has_fixed_port_id and b.has_fixed_port_id:
                    if a.fixed_port_id == b.fixed_port_id:
                        raise FixedPortIDCollisionError(
                            'The fixed port ID of this definition is also used in %r' % b.source_file_path,
                            path=a.source_file_path
                        )


def _ensure_minor_version_compatibility(types: typing.List[CompoundType]) -> None:
    by_name = defaultdict(list)  # type: typing.DefaultDict[str, typing.List[CompoundType]]
    for t in types:
        by_name[t.full_name].append(t)

    for name, definitions in by_name.items():
        by_major = defaultdict(list)    # type: typing.DefaultDict[int, typing.List[CompoundType]]
        for t in definitions:
            by_major[t.version.major].append(t)

        for subject_to_check in by_major.values():
            _logger.debug('Minor version compatibility check amongst: %s', [str(x) for x in subject_to_check])
            for a in subject_to_check:
                for b in subject_to_check:
                    if a is b:
                        continue

                    assert a.version.major == b.version.major
                    assert a.full_name == b.full_name

                    # Version collision
                    if a.version.minor == b.version.minor:
                        raise MultipleDefinitionsUnderSameVersionError(
                            'This definition shares its version number with %r' % b.source_file_path,
                            path=a.source_file_path
                        )

                    # Must be of the same kind: both messages or both services
                    if isinstance(a, ServiceType) != isinstance(b, ServiceType):
                        raise VersionsOfDifferentKindError(
                            'This definition is not of the same kind as %r' % b.source_file_path,
                            path=a.source_file_path
                        )

                    # Must be bit-compatible
                    if isinstance(a, ServiceType):
                        assert isinstance(b, ServiceType)
                        ok = a.request_type.is_mutually_bit_compatible_with(b.request_type) and \
                            a.response_type.is_mutually_bit_compatible_with(b.response_type)
                    else:
                        ok = a.compute_bit_length_values() == b.compute_bit_length_values()

                    if not ok:
                        raise MinorVersionsNotBitCompatibleError(
                            'This definition is not bit-compatible with %r' % b.source_file_path,
                            path=a.source_file_path
                        )

                    # Must use either the same RPID, or the older one should not have an RPID
                    if a.has_fixed_port_id == b.has_fixed_port_id:
                        if a.fixed_port_id != b.fixed_port_id:
                            raise MinorVersionFixedPortIDError(
                                'Different fixed port ID values under the same version %r' % b.source_file_path,
                                path=a.source_file_path
                            )
                    else:
                        must_have = a if a.version.minor > b.version.minor else b
                        if not must_have.has_fixed_port_id:
                            raise MinorVersionFixedPortIDError(
                                'Fixed port ID cannot be removed under the same major version',
                                path=must_have.source_file_path
                            )


def _ensure_no_nested_root_namespaces(directories: typing.Iterable[str]) -> None:
    directories = list(sorted([str(os.path.abspath(x)) for x in set(directories)]))
    for a in directories:
        for b in directories:
            if (a != b) and a.startswith(b):
                raise NestedRootNamespaceError(
                    'The following namespace is nested inside this one, which is not permitted: %r' % a,
                    path=b
                )


def _ensure_no_namespace_name_collisions(directories: typing.Iterable[str]) -> None:
    def get_namespace_name(d: str) -> str:
        return os.path.split(d)[-1]

    directories = list(sorted([str(os.path.abspath(x)) for x in set(directories)]))
    for a in directories:
        for b in directories:
            if (a != b) and get_namespace_name(a) == get_namespace_name(b):
                raise NamespaceNameCollisionError('The name of this namespace conflicts with %r' % b,
                                                  path=a)


def _construct_dsdl_definitions_from_namespace(root_namespace_path: str) -> typing.List[DSDLDefinition]:
    """
    Accepts a directory path, returns a list of abstract DSDL file representations.
    Those can be fed to the actual DSDL parser later.
    """
    def on_walk_error(os_ex: OSError) -> None:
        raise os_ex     # pragma: no cover

    walker = os.walk(root_namespace_path,
                     onerror=on_walk_error,
                     followlinks=True)

    source_file_paths = []  # type: typing.List[str]
    for root, _dirnames, filenames in walker:
        for filename in fnmatch.filter(filenames, DSDL_FILE_GLOB):
            source_file_paths.append(os.path.join(root, filename))

    _logger.debug('DSDL files in the namespace dir %r are listed below:', root_namespace_path)
    for a in source_file_paths:
        _logger.debug(_LOG_LIST_ITEM_PREFIX + a)

    output = []  # type: typing.List[DSDLDefinition]
    for fp in source_file_paths:
        dsdl_def = DSDLDefinition(fp, root_namespace_path)
        output.append(dsdl_def)

    return output


def _unittest_dsdl_definition_constructor() -> None:
    import tempfile
    from .dsdl_definition import FileNameFormatError

    directory = tempfile.TemporaryDirectory()
    root_ns_dir = os.path.join(directory.name, 'foo')

    os.mkdir(root_ns_dir)
    os.mkdir(os.path.join(root_ns_dir, 'nested'))

    def touchy(relative_path: str) -> None:
        p = os.path.join(root_ns_dir, relative_path)
        os.makedirs(os.path.dirname(p), exist_ok=True)
        with open(p, 'w') as f:
            f.write('# TEST TEXT')

    def discard(relative_path: str) -> None:
        os.unlink(os.path.join(root_ns_dir, relative_path))

    touchy('123.Qwerty.123.234.uavcan')
    touchy('nested/2.Asd.21.32.uavcan')
    touchy('nested/Foo.32.43.uavcan')

    dsdl_defs = _construct_dsdl_definitions_from_namespace(root_ns_dir)
    print(dsdl_defs)
    lut = {x.full_name: x for x in dsdl_defs}    # type: typing.Dict[str, DSDLDefinition]
    assert len(lut) == 3

    assert str(lut['foo.Qwerty']) == repr(lut['foo.Qwerty'])
    assert str(lut['foo.Qwerty']) == \
        "DSDLDefinition(name='foo.Qwerty', version=Version(major=123, minor=234), fixed_port_id=123, " \
        "file_path='%s')" % lut['foo.Qwerty'].file_path

    assert str(lut['foo.nested.Foo']) == \
        "DSDLDefinition(name='foo.nested.Foo', version=Version(major=32, minor=43), fixed_port_id=None, " \
        "file_path='%s')" % lut['foo.nested.Foo'].file_path

    t = lut['foo.Qwerty']
    assert t.file_path == os.path.join(root_ns_dir, '123.Qwerty.123.234.uavcan')
    assert t.has_fixed_port_id
    assert t.fixed_port_id == 123
    assert t.text == '# TEST TEXT'
    assert t.version.major == 123
    assert t.version.minor == 234
    assert t.name_components == ['foo', 'Qwerty']
    assert t.short_name == 'Qwerty'
    assert t.root_namespace == 'foo'
    assert t.full_namespace == 'foo'

    t = lut['foo.nested.Asd']
    assert t.file_path == os.path.join(root_ns_dir, 'nested', '2.Asd.21.32.uavcan')
    assert t.has_fixed_port_id
    assert t.fixed_port_id == 2
    assert t.text == '# TEST TEXT'
    assert t.version.major == 21
    assert t.version.minor == 32
    assert t.name_components == ['foo', 'nested', 'Asd']
    assert t.short_name == 'Asd'
    assert t.root_namespace == 'foo'
    assert t.full_namespace == 'foo.nested'

    t = lut['foo.nested.Foo']
    assert t.file_path == os.path.join(root_ns_dir, 'nested', 'Foo.32.43.uavcan')
    assert not t.has_fixed_port_id
    assert t.fixed_port_id is None
    assert t.text == '# TEST TEXT'
    assert t.version.major == 32
    assert t.version.minor == 43
    assert t.name_components == ['foo', 'nested', 'Foo']
    assert t.short_name == 'Foo'
    assert t.root_namespace == 'foo'
    assert t.full_namespace == 'foo.nested'

    touchy('nested/Malformed.MAJOR.MINOR.uavcan')
    try:
        _construct_dsdl_definitions_from_namespace(root_ns_dir)
    except FileNameFormatError as ex:
        print(ex)
        discard('nested/Malformed.MAJOR.MINOR.uavcan')
    else:       # pragma: no cover
        assert False

    touchy('nested/NOT_A_NUMBER.Malformed.1.0.uavcan')
    try:
        _construct_dsdl_definitions_from_namespace(root_ns_dir)
    except FileNameFormatError as ex:
        print(ex)
        discard('nested/NOT_A_NUMBER.Malformed.1.0.uavcan')
    else:       # pragma: no cover
        assert False

    touchy('nested/Malformed.uavcan')
    try:
        _construct_dsdl_definitions_from_namespace(root_ns_dir)
    except FileNameFormatError as ex:
        print(ex)
        discard('nested/Malformed.uavcan')
    else:       # pragma: no cover
        assert False

    _construct_dsdl_definitions_from_namespace(root_ns_dir)  # making sure all errors are cleared

    touchy('nested/super.bad/Unreachable.1.0.uavcan')
    try:
        _construct_dsdl_definitions_from_namespace(root_ns_dir)
    except FileNameFormatError as ex:
        print(ex)
    else:       # pragma: no cover
        assert False

    try:
        _construct_dsdl_definitions_from_namespace(root_ns_dir + '/nested/super.bad')
    except FileNameFormatError as ex:
        print(ex)
    else:       # pragma: no cover
        assert False

    discard('nested/super.bad/Unreachable.1.0.uavcan')



#
# Copyright (C) 2018-2019  UAVCAN Development Team  <uavcan.org>
# This software is distributed under the terms of the MIT License.
#

import re
import typing
import logging
import inspect
import operator
from .parse_error import ParseError, InternalError, InvalidDefinitionError
from .dsdl_definition import DSDLDefinition
from .data_type import BooleanType, SignedIntegerType, UnsignedIntegerType, FloatType, VoidType, DataType
from .data_type import ArrayType, StaticArrayType, DynamicArrayType, CompoundType, UnionType, StructureType
from .data_type import ServiceType, Attribute, Field, PaddingField, Constant, PrimitiveType
from .data_type import TypeParameterError, InvalidFixedPortIDError
from .regular_grammar_matcher import RegularGrammarMatcher, InvalidGrammarError, GrammarConstructHandler
from .port_id_ranges import is_valid_regulated_subject_id, is_valid_regulated_service_id


class DSDLSyntaxError(InvalidDefinitionError):
    pass


class SemanticError(InvalidDefinitionError):
    pass


class AssertionCheckFailureError(SemanticError):
    pass


class UndefinedDataTypeError(SemanticError):
    pass


# Arguments: emitting definition, line number, value to print
# The lines are numbered starting from one
PrintDirectiveOutputHandler = typing.Callable[[DSDLDefinition, int, typing.Any], None]


class ConfigurationOptions:
    def __init__(self) -> None:
        self.print_handler = None                       # type: typing.Optional[PrintDirectiveOutputHandler]
        self.allow_unregulated_fixed_port_id = False
        self.skip_assertion_checks = False


_GrammarRule = typing.NamedTuple('GrammarRule', [
    ('regexp', str),
    ('handler', GrammarConstructHandler)
])


class _ServiceResponseMarkerPlaceholder:
    pass


class _DirectivePlaceholder:
    def __init__(self,
                 directive: str,
                 expression: typing.Optional[str]) -> None:
        self.directive = directive
        self.expression = expression


_COMPOUND_ATTRIBUTE_TYPE_REGEXP = r'((?:[a-zA-Z_][a-zA-Z0-9_]*?\.)+?)(\d{1,3})\.(\d{1,3})$'


class _AttributeCollection:
    class _PostponedExpression:
        def __init__(self,
                     next_attribute_index: int,
                     expression_text: str,
                     validator: typing.Callable[[typing.Any], None]):
            self.next_attribute_index = int(next_attribute_index)
            self.expression_text = str(expression_text)
            self.validator = validator
            _logger.debug('Registering new postponed assertion check before attribute #%d: %s',
                          self.next_attribute_index, self.expression_text)

    def __init__(self) -> None:
        self.attributes = []    # type: typing.List[Attribute]
        self.is_union = False
        self._expressions = []  # type: typing.List['_AttributeCollection._PostponedExpression']

    def add_postponed_expression(self,
                                 expression_text: str,
                                 validator: typing.Callable[[typing.Any], None]) -> None:
        self._expressions.append(self._PostponedExpression(len(self.attributes),
                                                           expression_text,
                                                           validator))

    def execute_postponed_expressions(self, data_type: typing.Union[StructureType, UnionType]) -> None:
        for pe in self._expressions:
            offset = _OffsetValue(data_type, pe.next_attribute_index)
            result = _evaluate_expression(pe.expression_text,
                                          _offset_=offset)
            pe.validator(result)


_logger = logging.getLogger(__name__)


def parse_definition(definition:            DSDLDefinition,
                     lookup_definitions:    typing.Sequence[DSDLDefinition],
                     configuration_options: ConfigurationOptions) -> CompoundType:
    _logger.info('Parsing definition %r', definition)

    attribute_collections = [_AttributeCollection()]
    is_deprecated = False

    def mark_as_union() -> None:
        if len(attribute_collections[-1].attributes) > 0:
            raise SemanticError('Union directive must be placed before the first attribute declaration')

        attribute_collections[-1].is_union = True

    def mark_deprecated() -> None:
        nonlocal is_deprecated
        if (len(attribute_collections) > 1) or (len(attribute_collections[-1].attributes) > 0):
            raise SemanticError('Deprecated directive must be placed near the beginning of the type definition')

        is_deprecated = True

    def assert_expression(expression: str) -> None:
        def validator(result: typing.Any) -> None:
            if isinstance(result, bool):
                if not result:
                    raise AssertionCheckFailureError('Assertion check failed on %r' % expression)
            else:
                raise SemanticError('Assertion check expressions must yield a boolean; %r yields %r' %
                                    (expression, result))

        if not configuration_options.skip_assertion_checks:
            attribute_collections[-1].add_postponed_expression(expression, validator)

    def make_print_expression_handler(ln: int) -> typing.Callable[[str], None]:
        # An extra closure is needed to capture the line number
        def fun(expression: str) -> None:
            def validator(result: typing.Any) -> None:
                _logger.info('@print: %s:%d: %r' % (definition.file_path, ln, result))
                if configuration_options.print_handler:
                    configuration_options.print_handler(definition, ln, result)

            attribute_collections[-1].add_postponed_expression(expression, validator)

        return fun

    try:
        for line_number, output in _evaluate(definition,
                                             list(lookup_definitions),
                                             configuration_options):
            if isinstance(output, Attribute):
                attribute_collections[-1].attributes.append(output)

            elif isinstance(output, _ServiceResponseMarkerPlaceholder):
                if len(attribute_collections) > 1:
                    raise SemanticError('Duplicate service response marker')
                else:
                    attribute_collections.append(_AttributeCollection())
                    assert len(attribute_collections) == 2

            elif isinstance(output, _DirectivePlaceholder):
                try:
                    handler = {
                        'union':      mark_as_union,
                        'deprecated': mark_deprecated,
                        'assert':     assert_expression,
                        'print':      make_print_expression_handler(line_number),
                    }[output.directive]
                except KeyError:
                    raise SemanticError('Unknown directive: %r' % output.directive)
                else:
                    num_parameters = len(inspect.signature(handler).parameters)  # type: ignore
                    assert 0 <= num_parameters <= 1, 'Invalid directive handler'
                    expression_required = num_parameters > 0

                    if expression_required and not output.expression:
                        raise SemanticError('Directive %r requires an expression' % output.directive)

                    if output.expression and not expression_required:
                        raise SemanticError('Directive %r does not expect an expression' % output.directive)

                    _logger.debug('Executing directive %r with expression %r',
                                  output.directive, output.expression)
                    if expression_required:
                        assert output.expression
                        handler(output.expression)  # type: ignore
                    else:
                        handler()  # type: ignore

            elif output is None:
                pass

            else:   # pragma: no cover
                assert False, 'Unexpected output'

    except ParseError as ex:  # pragma: no cover
        ex.set_error_location_if_unknown(path=definition.file_path)
        raise

    except Exception as ex:  # pragma: no cover
        raise InternalError(culprit=ex, path=definition.file_path)

    try:
        if len(attribute_collections) == 1:
            ac = attribute_collections[-1]
            if ac.is_union:
                tout = UnionType(name=definition.full_name,
                                 version=definition.version,
                                 attributes=ac.attributes,
                                 deprecated=is_deprecated,
                                 fixed_port_id=definition.fixed_port_id,
                                 source_file_path=definition.file_path)    # type: CompoundType
            else:
                tout = StructureType(name=definition.full_name,
                                     version=definition.version,
                                     attributes=ac.attributes,
                                     deprecated=is_deprecated,
                                     fixed_port_id=definition.fixed_port_id,
                                     source_file_path=definition.file_path)

            assert isinstance(tout, (StructureType, UnionType))
            ac.execute_postponed_expressions(tout)
        else:
            req, res = attribute_collections       # type: _AttributeCollection, _AttributeCollection
            tout = ServiceType(name=definition.full_name,
                               version=definition.version,
                               request_attributes=req.attributes,
                               response_attributes=res.attributes,
                               request_is_union=req.is_union,
                               response_is_union=res.is_union,
                               deprecated=is_deprecated,
                               fixed_port_id=definition.fixed_port_id,
                               source_file_path=definition.file_path)

            assert isinstance(tout, ServiceType)
            for ac, dt in [
                (req, tout.request_type),
                (res, tout.response_type),
            ]:
                assert isinstance(dt, (StructureType, UnionType))
                ac.execute_postponed_expressions(dt)

        # Regulated fixed port ID check
        if not configuration_options.allow_unregulated_fixed_port_id:
            port_id = tout.fixed_port_id
            if port_id is not None:
                f = is_valid_regulated_service_id if isinstance(tout, ServiceType) else is_valid_regulated_subject_id
                if not f(port_id, tout.root_namespace):
                    raise InvalidFixedPortIDError('Regulated port ID %r is not valid. '
                                                  'Consider using allow_unregulated_fixed_port_id.' % port_id)

        assert isinstance(tout, CompoundType)
        return tout
    except TypeParameterError as ex:
        raise SemanticError(str(ex), path=definition.file_path)
    except ParseError as ex:  # pragma: no cover
        ex.set_error_location_if_unknown(path=definition.file_path)
        raise


def _evaluate(definition: DSDLDefinition,
              lookup_definitions: typing.List[DSDLDefinition],
              configuration_options: ConfigurationOptions) -> \
        typing.Iterable[typing.Tuple[int, typing.Union[None,
                                                       Attribute,
                                                       _ServiceResponseMarkerPlaceholder,
                                                       _DirectivePlaceholder]]]:
    ns = definition.full_namespace

    grammar = RegularGrammarMatcher()
    grammar.add_rule(*_make_scalar_field_rule(ns, lookup_definitions, configuration_options))
    grammar.add_rule(*_make_array_field_rule(ns, lookup_definitions, configuration_options))
    grammar.add_rule(*_make_constant_rule(ns, lookup_definitions, configuration_options))
    grammar.add_rule(*_make_padding_rule())
    grammar.add_rule(*_make_service_response_marker_rule())
    grammar.add_rule(*_make_directive_rule())
    grammar.add_rule(*_make_empty_rule())

    for line_index, line_text in enumerate(definition.text.splitlines(keepends=False)):
        line_number = line_index + 1
        try:
            output = grammar.match(line_text)
            yield line_number, output
        except InvalidGrammarError:
            raise DSDLSyntaxError('Syntax error',
                                  path=definition.file_path,
                                  line=line_number)
        except TypeParameterError as ex:
            raise SemanticError(str(ex),
                                path=definition.file_path,
                                line=line_number)
        except ParseError as ex:  # pragma: no cover
            ex.set_error_location_if_unknown(path=definition.file_path,
                                             line=line_number)
            raise
        except Exception as ex:  # pragma: no cover
            raise InternalError(culprit=ex,
                                path=definition.file_path,
                                line=line_number)


def _make_scalar_field_rule(referer_namespace:  str,
                            lookup_definitions: typing.List[DSDLDefinition],
                            configuration_options: ConfigurationOptions) -> _GrammarRule:
    def constructor(cast_mode: typing.Optional[str],
                    type_name: str,
                    field_name: str) -> Field:
        t = _construct_type(referer_namespace=referer_namespace,
                            cast_mode=cast_mode,
                            type_name=type_name,
                            lookup_definitions=lookup_definitions,
                            configuration_options=configuration_options)
        return Field(t, field_name)

    return _GrammarRule(r'\s*(?:(saturated|truncated)\s+)?'         # Cast mode
                        r'([a-zA-Z_][a-zA-Z0-9_\.]*)\s+'            # Type name
                        r'([a-zA-Z_][a-zA-Z0-9_]*)'                 # Field name
                        r'\s*(?:#.*)?$',                            # End of the line
                        constructor)


def _make_array_field_rule(referer_namespace:  str,
                           lookup_definitions: typing.List[DSDLDefinition],
                           configuration_options: ConfigurationOptions) -> _GrammarRule:
    def constructor(cast_mode: typing.Optional[str],
                    type_name: str,
                    mode_specifier: typing.Optional[str],
                    size_specifier: str,
                    field_name: str) -> Field:
        e = _construct_type(referer_namespace=referer_namespace,
                            cast_mode=cast_mode,
                            type_name=type_name,
                            lookup_definitions=lookup_definitions,
                            configuration_options=configuration_options)
        # The size specifier is guaranteed to be a valid integer, see the regular expression
        size = int(size_specifier, 0)

        if not mode_specifier:
            t = StaticArrayType(element_type=e, size=size)              # type: ArrayType
        elif mode_specifier == '<':
            t = DynamicArrayType(element_type=e, max_size=size - 1)
        elif mode_specifier == '<=':
            t = DynamicArrayType(element_type=e, max_size=size)
        else:   # pragma: no cover
            raise InternalError('Choo choo bitches')

        return Field(t, field_name)

    return _GrammarRule(r'\s*(?:(saturated|truncated)\s+)?'         # Cast mode
                        r'([a-zA-Z_][a-zA-Z0-9_\.]*)\s*'            # Type name
                        r'\[\s*(<=?)?\s*(\d+)\s*\]\s+'              # Mode/size specifier
                        r'([a-zA-Z_][a-zA-Z0-9_]*)'                 # Field name
                        r'\s*(?:#.*)?$',                            # End of the line
                        constructor)


def _make_constant_rule(referer_namespace:  str,
                        lookup_definitions: typing.List[DSDLDefinition],
                        configuration_options: ConfigurationOptions) -> _GrammarRule:
    def constructor(cast_mode: typing.Optional[str],
                    type_name: str,
                    constant_name: str,
                    initialization_expression: str) -> Constant:
        t = _construct_type(referer_namespace=referer_namespace,
                            cast_mode=cast_mode,
                            type_name=type_name,
                            lookup_definitions=lookup_definitions,
                            configuration_options=configuration_options)
        try:
            value = _evaluate_expression(initialization_expression)
        except SyntaxError as ex:
            raise DSDLSyntaxError('Malformed initialization expression: %s' % ex)
        except Exception as ex:
            raise SemanticError('Could not evaluate the constant initialization expression: %s' % ex)

        return Constant(data_type=t,
                        name=constant_name,
                        value=value,
                        initialization_expression=initialization_expression)

    return _GrammarRule(r'\s*(?:(saturated|truncated)\s+)?'         # Cast mode
                        r'([a-zA-Z_][a-zA-Z0-9_\.]*)\s+'            # Type name
                        r'([a-zA-Z_][a-zA-Z0-9_]*)'                 # Constant name
                        r'\s*=\s*'                                  # Assignment
                        r"((?:'[^']')|(?:[+\-\.0-9a-zA-Z_]+))"      # Initialization expression
                        r'\s*(?:#.*)?$',                            # End of the line
                        constructor)


def _make_padding_rule() -> _GrammarRule:
    return _GrammarRule(r'\s*void(\d\d?)\s*(?:#.*)?$',
                        lambda bw: PaddingField(VoidType(int(bw))))


def _make_service_response_marker_rule() -> _GrammarRule:
    return _GrammarRule(r'\s*---\s*(?:#.*)?$',
                        lambda: _ServiceResponseMarkerPlaceholder())


def _make_directive_rule() -> _GrammarRule:
    # The fact that the DSDL grammar is so simple allows us to get by with ridiculously simple expressions here.
    # We just take everything between the directive itself and either the end of the line or the first comment
    # and treat it as the expression (to be handled later). Strings are not allowed inside the expression,
    # which simplifies handling greatly.
    return _GrammarRule(r'\s*@([a-zA-Z0-9_]+)\s*([^#]+?)?\s*(?:#.*)?$',
                        lambda d, e: _DirectivePlaceholder(d, e))


def _make_empty_rule() -> _GrammarRule:
    return _GrammarRule(r'\s*(?:#.*)?$', lambda: None)


def _construct_type(referer_namespace:  str,
                    cast_mode:          typing.Optional[str],
                    type_name:          str,
                    lookup_definitions: typing.List[DSDLDefinition],
                    configuration_options: ConfigurationOptions) -> DataType:
    assert referer_namespace == referer_namespace.strip().strip(CompoundType.NAME_COMPONENT_SEPARATOR).strip()

    def get_cast_mode() -> PrimitiveType.CastMode:
        return {
            'truncated': PrimitiveType.CastMode.TRUNCATED,
            'saturated': PrimitiveType.CastMode.SATURATED,
            None:        PrimitiveType.CastMode.SATURATED,
        }[cast_mode]

    def construct_compound(name: str, v_major: int, v_minor: int) -> CompoundType:
        if cast_mode is not None:
            raise SemanticError('Cast mode cannot be specified for compound data types')

        if CompoundType.NAME_COMPONENT_SEPARATOR not in name:
            # Namespace not specified, this means that we're using relative reference
            absolute_name = CompoundType.NAME_COMPONENT_SEPARATOR.join([referer_namespace, name])  # type: str
            _logger.debug('Relative reference: %r --> %r', name, absolute_name)
            name = absolute_name

        matching_name = list(filter(lambda x: x.full_name == name, lookup_definitions))
        if not matching_name:
            raise UndefinedDataTypeError('No type named ' + name)

        matching_major = list(filter(lambda x: x.version.major == v_major, matching_name))
        if not matching_major:
            raise UndefinedDataTypeError('No suitable major version of %r could be found. '
                                         'Requested version %d, found: %r' % (name, v_major, matching_name))

        matching_minor = list(filter(lambda x: x.version.minor == v_minor, matching_major))
        if not matching_minor:
            raise UndefinedDataTypeError('No suitable minor version of %r could be found. '
                                         'Requested minor version %d, found: %r' %
                                         (name, v_minor, matching_major))

        if len(matching_minor) != 1:    # pragma: no cover
            raise InternalError('Unexpected ambiguity: %r' % matching_minor)

        definition = matching_minor[0]

        # Remove all versions of the same type from lookup definitions to prevent circular dependencies
        lookup_definitions_without_circular_dependencies = [
            x for x in lookup_definitions if x.full_name != definition.full_name
        ]

        return parse_definition(definition,
                                lookup_definitions_without_circular_dependencies,
                                configuration_options)

    g = RegularGrammarMatcher()
    g.add_rule(r'bool$', lambda: BooleanType(get_cast_mode()))
    g.add_rule(r'void(\d\d?)$', lambda bw: VoidType(int(bw)))
    g.add_rule(r'float(\d\d?)$', lambda bw: FloatType(int(bw), get_cast_mode()))
    g.add_rule(r'int(\d\d?)$', lambda bw: SignedIntegerType(int(bw), get_cast_mode()))
    g.add_rule(r'uint(\d\d?)$', lambda bw: UnsignedIntegerType(int(bw), get_cast_mode()))
    g.add_rule(_COMPOUND_ATTRIBUTE_TYPE_REGEXP,
               lambda name, v_mj, v_mn: construct_compound(name.strip('.'), int(v_mj), int(v_mn)))
    try:
        t = g.match(type_name)
    except InvalidGrammarError:
        raise DSDLSyntaxError('Invalid type declaration: ' + type_name)
    else:
        assert isinstance(t, DataType)
        _logger.debug('Type constructed successfully: %r --> %r', type_name, t)
        return t


def _evaluate_expression(expression: str, **context: typing.Any) -> typing.Any:
    env = {
        'locals': None,
        'globals': None,
        '__builtins__': None,
        'true': True,
        'false': False,
    }
    env.update(context)
    return eval(expression, env)


class _OffsetValue:
    def __init__(self,
                 data_type: CompoundType,
                 next_attribute_index: int):
        self._data_type = data_type

        self._next_field_index = 0
        for i, a in enumerate(self._data_type.attributes[:next_attribute_index]):
            if isinstance(a, Field):
                self._next_field_index += 1

        assert next_attribute_index >= self._next_field_index
        _logger.debug('Index conversion: attribute #%d --> field #%d for %r',
                      next_attribute_index, self._next_field_index, data_type)

        self._set_cache = None      # type: typing.Optional[typing.Set[int]]

    @property
    def _set(self) -> typing.Set[int]:
        # We're using lazy evaluation because not every expression uses the offset value
        if self._set_cache is None:
            if self._next_field_index >= len(self._data_type.fields):
                self._set_cache = set(self._data_type.compute_bit_length_values())
            else:
                if isinstance(self._data_type, StructureType):
                    self._set_cache = self._data_type.get_field_offset_values(field_index=self._next_field_index)
                elif isinstance(self._data_type, UnionType):
                    raise SemanticError('Inter-field min/max offset is not defined for unions')
                else:   # pragma: no cover
                    assert False, 'Ill-defined offset'

        assert len(self._set_cache) > 0, 'Empty BLV sets are forbidden'
        return self._set_cache

    @property
    def min(self) -> int:
        return min(self._set)   # Can be optimized for the case when next_field_index == len(fields)

    @property
    def max(self) -> int:
        return max(self._set)   # Can be optimized for the case when next_field_index == len(fields)

    def _do_elementwise(self,
                        element_operator: typing.Callable[[int, int], int],
                        right_hand_operand: int) -> typing.Set[int]:
        if isinstance(right_hand_operand, int):
            return set(map(lambda x: element_operator(x, right_hand_operand), self._set))
        else:
            raise SemanticError('Invalid operand %r' % right_hand_operand)

    def __eq__(self, other: typing.Any) -> bool:
        if isinstance(other, set):
            return other == self._set
        elif isinstance(other, _OffsetValue):
            return self._set == other._set
        else:
            raise SemanticError('Offset cannot be compared against %r' % other)

    def __mod__(self, other: int) -> typing.Set[int]:
        return self._do_elementwise(operator.mod, other)

    def __add__(self, other: int) -> typing.Set[int]:
        return self._do_elementwise(operator.add, other)

    def __sub__(self, other: int) -> typing.Set[int]:
        return self._do_elementwise(operator.sub, other)

    def __mul__(self, other: int) -> typing.Set[int]:
        return self._do_elementwise(operator.mul, other)

    def __truediv__(self, other: int) -> typing.Set[int]:
        """Floor division using the true division syntax"""
        return self._do_elementwise(operator.floordiv, other)

    __rmul__ = __mul__
    __radd__ = __add__

    # We can't rely on functools.total_order because we use unconventional elementwise operators
    def __lt__(self, other: int) -> bool:
        return all(self._do_elementwise(operator.lt, other))

    def __le__(self, other: int) -> bool:
        return all(self._do_elementwise(operator.le, other))

    def __gt__(self, other: int) -> bool:
        return all(self._do_elementwise(operator.gt, other))

    def __ge__(self, other: int) -> bool:
        return all(self._do_elementwise(operator.ge, other))

    def __str__(self) -> str:
        return str(self._set or '{}')

    __repr__ = __str__


def _unittest_regexp() -> None:
    def validate(pattern: str,
                 text: str,
                 expected_output: typing.Optional[typing.Sequence[typing.Optional[str]]]) -> None:
        match = re.match(pattern, text)
        if expected_output is not None:
            assert match
            assert list(expected_output) == list(match.groups())
        else:
            assert match is None

    validate(_make_scalar_field_rule('', [], ConfigurationOptions()).regexp,
             'saturated uint8 value',
             ('saturated', 'uint8', 'value'))

    # This is not the intended behavior, but a side effect of the regular grammar we're using
    validate(_make_scalar_field_rule('', [], ConfigurationOptions()).regexp,
             'saturated uint8',
             (None, 'saturated', 'uint8'))

    validate(_make_scalar_field_rule('', [], ConfigurationOptions()).regexp,
             ' namespace.nested.TypeName.0.1  _0  # comment',
             (None, 'namespace.nested.TypeName.0.1', '_0'))

    validate(_make_array_field_rule('', [], ConfigurationOptions()).regexp,
             'namespace.nested.TypeName.0.1[123] _0',
             (None, 'namespace.nested.TypeName.0.1', None, '123', '_0'))

    validate(_make_array_field_rule('', [], ConfigurationOptions()).regexp,
             '  namespace.nested.TypeName.0.1  [   123  ]  _# comment',
             (None, 'namespace.nested.TypeName.0.1', None, '123', '_'))

    validate(_make_array_field_rule('', [], ConfigurationOptions()).regexp,
             'truncated type[<123] _0',
             ('truncated', 'type', '<', '123', '_0'))

    validate(_make_array_field_rule('', [], ConfigurationOptions()).regexp,
             'truncated type[<=123] _0',
             ('truncated', 'type', '<=', '123', '_0'))

    validate(_make_array_field_rule('', [], ConfigurationOptions()).regexp,
             ' truncated type  [ <= 123 ] _0',
             ('truncated', 'type', '<=', '123', '_0'))

    validate(_make_padding_rule().regexp, 'void1', ('1',))
    validate(_make_padding_rule().regexp, ' void64 ', ('64',))

    validate(_make_constant_rule('', [], ConfigurationOptions()).regexp,
             'uint8 NAME = 123',
             (None, 'uint8', 'NAME', '123'))

    validate(_make_constant_rule('', [], ConfigurationOptions()).regexp,
             'uint8 NAME = +123.456e+123',
             (None, 'uint8', 'NAME', '+123.456e+123'))

    validate(_make_constant_rule('', [], ConfigurationOptions()).regexp,
             'uint8 NAME = -0xabcdef',
             (None, 'uint8', 'NAME', '-0xabcdef'))

    validate(_make_constant_rule('', [], ConfigurationOptions()).regexp,
             'uint8 NAME = -0o123456',
             (None, 'uint8', 'NAME', '-0o123456'))

    validate(_make_constant_rule('', [], ConfigurationOptions()).regexp,
             ' uint8 NAME=123#comment',
             (None, 'uint8', 'NAME', '123'))

    validate(_make_constant_rule('', [], ConfigurationOptions()).regexp,
             " uint8 NAME='#'",
             (None, 'uint8', 'NAME', "'#'"))

    validate(_make_constant_rule('', [], ConfigurationOptions()).regexp,
             "\tuint8 NAME = '#'# comment",
             (None, 'uint8', 'NAME', "'#'"))

    validate(_make_directive_rule().regexp,
             "@directive",
             ('directive', None))

    validate(_make_directive_rule().regexp,
             " @directive # hello world",
             ('directive', None))

    validate(_make_directive_rule().regexp,
             " @directive a + b == c # hello world",
             ('directive', 'a + b == c'))

    validate(_make_directive_rule().regexp,
             " @directive a+b==c#hello world",
             ('directive', 'a+b==c'))

    validate(_make_directive_rule().regexp,
             " @directive a+b==c",
             ('directive', 'a+b==c'))

    validate(_make_service_response_marker_rule().regexp, "---", ())
    validate(_make_service_response_marker_rule().regexp, "\t---", ())
    validate(_make_service_response_marker_rule().regexp, "---  ", ())
    validate(_make_service_response_marker_rule().regexp, " ---  # whatever", ())
    validate(_make_service_response_marker_rule().regexp, "---#whatever", ())

    re_empty = _make_empty_rule().regexp
    validate(re_empty, '', ())
    validate(re_empty, ' ', ())
    validate(re_empty, ' \t ', ())
    validate(re_empty, ' \t # Whatever! 987g13eh_bv-0o%e5tjkn rtgb-y562254-/*986+', ())
    validate(re_empty, '### Whatever! 987g13eh_bv-0o%e5tjkn rtgb-y562254-/*986+', ())
    validate(re_empty, '132', None)
    validate(re_empty, ' abc', None)
    validate(re_empty, ' "#" ', None)
    validate(re_empty, '"#"', None)
    validate(re_empty, "'#'", None)

    validate(_COMPOUND_ATTRIBUTE_TYPE_REGEXP,
             'uavcan.node.Heartbeat.1.2',
             ('uavcan.node.Heartbeat.', '1', '2'))

    validate(_COMPOUND_ATTRIBUTE_TYPE_REGEXP,
             'uavcan.node.Heartbeat.1',
             None)

    validate(_COMPOUND_ATTRIBUTE_TYPE_REGEXP,
             'uavcan.node.Heartbeat.1.',
             None)

    validate(_COMPOUND_ATTRIBUTE_TYPE_REGEXP,
             'Heartbeat.1.2',
             ('Heartbeat.', '1', '2'))

    validate(_COMPOUND_ATTRIBUTE_TYPE_REGEXP,
             'a1.123',
             None)

    validate(_COMPOUND_ATTRIBUTE_TYPE_REGEXP,
             'a1.123.234',
             ('a1.', '123', '234'))

    validate(_COMPOUND_ATTRIBUTE_TYPE_REGEXP,
             'uavcan.1node.Heartbeat.1.2',
             None)

    validate(_COMPOUND_ATTRIBUTE_TYPE_REGEXP,
             'uavcan.node..Heartbeat.1.2',
             None)

    validate(_COMPOUND_ATTRIBUTE_TYPE_REGEXP,
             'uavcan.node.Heartbeat..1.2',
             None)

    validate(_COMPOUND_ATTRIBUTE_TYPE_REGEXP,
             'uavcan.node.Heartbeat.1..2',
             None)

    validate(_COMPOUND_ATTRIBUTE_TYPE_REGEXP,
             'uavcan.node.Heartbeat.1.2.',
             None)

    validate(_COMPOUND_ATTRIBUTE_TYPE_REGEXP,
             'uavcan.node-Heartbeat.1.2',
             None)



#
# Copyright (C) 2018-2019  UAVCAN Development Team  <uavcan.org>
# This software is distributed under the terms of the MIT License.
#

MAX_SUBJECT_ID = 32767
MAX_SERVICE_ID = 511


_STANDARD_ROOT_NAMESPACE = 'uavcan'

_STANDARD_MESSAGES = 31744, 32767
_VENDOR_MESSAGES   = 28672, 29695

_STANDARD_SERVICES = 384, 511
_VENDOR_SERVICES   = 256, 319


def is_valid_regulated_subject_id(regulated_id: int,
                                  root_namespace: str) -> bool:
    is_standard = root_namespace.strip() == _STANDARD_ROOT_NAMESPACE
    applicable_range = _STANDARD_MESSAGES if is_standard else _VENDOR_MESSAGES
    return applicable_range[0] <= int(regulated_id) <= applicable_range[1]


def is_valid_regulated_service_id(regulated_id: int,
                                  root_namespace: str) -> bool:
    is_standard = root_namespace.strip() == _STANDARD_ROOT_NAMESPACE
    applicable_range = _STANDARD_SERVICES if is_standard else _VENDOR_SERVICES
    return applicable_range[0] <= int(regulated_id) <= applicable_range[1]


def _unittest_pid_ranges() -> None:
    assert is_valid_regulated_subject_id(regulated_id=29000, root_namespace='sirius_cybernetics_corp')
    assert not is_valid_regulated_subject_id(regulated_id=29000, root_namespace='uavcan')
    assert is_valid_regulated_subject_id(regulated_id=32000, root_namespace='uavcan')
    assert not is_valid_regulated_subject_id(regulated_id=32000, root_namespace='sirius_cybernetics_corp')
    assert not is_valid_regulated_subject_id(regulated_id=30000, root_namespace='uavcan')
    assert not is_valid_regulated_subject_id(regulated_id=30000, root_namespace='sirius_cybernetics_corp')

    assert is_valid_regulated_service_id(regulated_id=260, root_namespace='sirius_cybernetics_corp')
    assert not is_valid_regulated_service_id(regulated_id=260, root_namespace='uavcan')
    assert is_valid_regulated_service_id(regulated_id=400, root_namespace='uavcan')
    assert not is_valid_regulated_service_id(regulated_id=400, root_namespace='sirius_cybernetics_corp')
    assert not is_valid_regulated_service_id(regulated_id=600, root_namespace='uavcan')
    assert not is_valid_regulated_service_id(regulated_id=600, root_namespace='sirius_cybernetics_corp')



#
# Copyright (C) 2018-2019  UAVCAN Development Team  <uavcan.org>
# This software is distributed under the terms of the MIT License.
#

import re
import typing
import logging


GrammarConstructHandler = typing.Union[
    typing.Callable[[], typing.Any],
    typing.Callable[[str], typing.Any],
    typing.Callable[[str, str], typing.Any],
    typing.Callable[[str, str, str], typing.Any],
    typing.Callable[[str, str, str, str], typing.Any],
    typing.Callable[[str, str, str, str, str], typing.Any],
    typing.Callable[[str, str, str, str, str, str], typing.Any],
]


_logger = logging.getLogger(__name__)


class InvalidGrammarError(ValueError):
    pass


class RegularGrammarMatcher:
    """
    Holds a collection of regular expression that together define a simple regular grammar.
    Can process text, matching the defined grammar rules against it; invokes a specified handler on first match
    and returns its output. If no match is found, raises InvalidGrammarError. The arguments of the handler are
    the captured strings, if any are specified; nothing otherwise.
    """

    def __init__(self) -> None:
        # static type not specified because mypy is malfunctioning on re.Pattern
        # noinspection Mypy
        self._rules = []        # type: ignore

    def add_rule(self,
                 regular_expression: str,
                 handler: GrammarConstructHandler) -> None:
        self._rules.append((re.compile(regular_expression), handler))

    def match(self, text: str) -> typing.Any:
        for regexp, handler in self._rules:
            match = re.match(regexp, text)
            if match:
                captured = match.groups()
                _logger.debug('Text %r produced %r matching this: %s', text, captured, regexp.pattern)
                return handler(*captured)

        raise InvalidGrammarError('Invalid grammar: %s' % text)


def _unittest_regular_grammar_matcher() -> None:
    from pytest import raises

    m = RegularGrammarMatcher()
    m.add_rule(r'float(\d\d?)$', lambda bits: 'F%d' % int(bits))
    m.add_rule(r'int(\d\d?)$', lambda bits: 'I%d' % int(bits))
    m.add_rule(r'uint(\d\d?)$', lambda bits: 'U%d' % int(bits))
    m.add_rule(r'void(\d\d?)$', lambda bits: 'V%d' % int(bits))
    m.add_rule(r'float(\d\d?)$', lambda _: None)  # Will never be invoked - consumed by previously defined
    m.add_rule(r'([a-zA-Z0-9_\.]+?)\.(\d+)(?:.(\d+))?$', lambda n, j, m: (n, int(j), None if m is None else int(m)))

    assert m.match('float64') == 'F64'
    assert m.match('int13') == 'I13'
    assert m.match('void52') == 'V52'

    with raises(InvalidGrammarError):
        m.match('no match')

    assert m.match('namespace.nested.Type.123.456') == ('namespace.nested.Type', 123, 456)
    assert m.match('namespace.nested.Type.123') == ('namespace.nested.Type', 123, None)
    with raises(InvalidGrammarError):
        m.match('namespace.nested.Type.')

    with raises(InvalidGrammarError):
        m.match('namespace.nested.Type.123.')



#!/usr/bin/env python3
#
# Copyright (C) 2018  UAVCAN Development Team  <uavcan.org>
# This software is distributed under the terms of the MIT License.
#
# Author: Pavel Kirienko <pavel.kirienko@zubax.com>
#

import sys
import setuptools

if int(setuptools.__version__.split('.')[0]) < 30:
    print('A newer version of setuptools is required. The current version does not support declarative config.',
          file=sys.stderr)
    sys.exit(1)

setuptools.setup()



#
# Copyright (C) 2018-2019  UAVCAN Development Team  <uavcan.org>
# This software is distributed under the terms of the MIT License.
#

import os
import typing
from .parse_error import InvalidDefinitionError
from .data_type import Version, CompoundType


class FileNameFormatError(InvalidDefinitionError):
    """
    Raised when a DSDL definition file is named incorrectly.
    """
    def __init__(self, text: str, path: str):
        super(FileNameFormatError, self).__init__(text=text, path=str(path))


class DSDLDefinition:
    """
    A DSDL type definition source abstracts the filesystem level details away, presenting a higher-level
    interface that operates solely on the level of type names, namespaces, fixed identifiers, and so on.
    Upper layers that operate on top of this abstraction do not concern themselves with the file system at all.
    """

    def __init__(self,
                 file_path: str,
                 root_namespace_path: str):
        # Normalizing the path and reading the definition text
        self._file_path = os.path.abspath(file_path)
        root_namespace_path = os.path.abspath(root_namespace_path)
        with open(self._file_path) as f:
            self._text = str(f.read())

        # Checking the sanity of the root directory path - can't contain separators
        if CompoundType.NAME_COMPONENT_SEPARATOR in os.path.split(root_namespace_path)[-1]:
            raise FileNameFormatError('Invalid namespace name', path=root_namespace_path)

        # Determining the relative path within the root namespace directory
        relative_path = str(os.path.join(os.path.split(root_namespace_path)[-1],
                                         os.path.relpath(self._file_path, root_namespace_path)))

        relative_directory, basename = [str(x) for x in os.path.split(relative_path)]   # type: str, str

        # Parsing the basename, e.g., 434.GetTransportStatistics.0.1.uavcan
        basename_components = basename.split('.')[:-1]
        str_fixed_port_id = None    # type: typing.Optional[str]
        if len(basename_components) == 4:
            str_fixed_port_id, short_name, str_major_version, str_minor_version = basename_components
        elif len(basename_components) == 3:
            short_name, str_major_version, str_minor_version = basename_components
        else:
            raise FileNameFormatError('Invalid file name', path=self._file_path)

        # Parsing the fixed port ID, if specified; None if not
        if str_fixed_port_id is not None:
            try:
                self._fixed_port_id = int(str_fixed_port_id)    # type: typing.Optional[int]
            except ValueError:
                raise FileNameFormatError('Could not parse the fixed port ID', path=self._file_path) from None
        else:
            self._fixed_port_id = None

        # Parsing the version numbers
        try:
            self._version = Version(major=int(str_major_version),
                                    minor=int(str_minor_version))
        except ValueError:
                raise FileNameFormatError('Could not parse the version numbers', path=self._file_path) from None

        # Finally, constructing the name
        namespace_components = list(relative_directory.strip(os.sep).split(os.sep))
        for nc in namespace_components:
            if CompoundType.NAME_COMPONENT_SEPARATOR in nc:
                raise FileNameFormatError('Invalid name for namespace component', path=self._file_path)

        self._name = CompoundType.NAME_COMPONENT_SEPARATOR.join(namespace_components + [str(short_name)])  # type: str

    @property
    def full_name(self) -> str:
        """The full name, e.g., uavcan.node.Heartbeat"""
        return self._name

    @property
    def name_components(self) -> typing.List[str]:
        """Components of the full name as a list, e.g., ['uavcan', 'node', 'Heartbeat']"""
        return self._name.split(CompoundType.NAME_COMPONENT_SEPARATOR)

    @property
    def short_name(self) -> str:
        """The last component of the full name, e.g., Heartbeat of uavcan.node.Heartbeat"""
        return self.name_components[-1]

    @property
    def full_namespace(self) -> str:
        """The full name without the short name, e.g., uavcan.node for uavcan.node.Heartbeat"""
        return str(CompoundType.NAME_COMPONENT_SEPARATOR.join(self.name_components[:-1]))

    @property
    def root_namespace(self) -> str:
        """The first component of the full name, e.g., uavcan of uavcan.node.Heartbeat"""
        return self.name_components[0]

    @property
    def text(self) -> str:
        """The source text in its raw unprocessed form (with comments, formatting intact, and everything)"""
        return self._text

    @property
    def version(self) -> Version:
        return self._version

    @property
    def fixed_port_id(self) -> typing.Optional[int]:
        """Either the fixed port ID as integer, or None if not defined for this type."""
        return self._fixed_port_id

    @property
    def has_fixed_port_id(self) -> bool:
        return self.fixed_port_id is not None

    @property
    def file_path(self) -> str:
        return self._file_path

    def __str__(self) -> str:
        return 'DSDLDefinition(name=%r, version=%r, fixed_port_id=%r, file_path=%r)' % \
            (self.full_name, self.version, self.fixed_port_id, self.file_path)

    __repr__ = __str__
